//Copyright (c) 1998 The Open Group
//
//Permission to use, copy, modify, and distribute this software and its documentation for
//any purpose and without fee is hereby granted, provided that the above copyright notice
//appears in all copies and that both the copyright notice and this permission notice
//appear in supporting documentation, and that the name of ("TOG") or The Open Group not
//be used in advertising or publicity pertaining to distribution of the software without
//specific, written prior permission.
//
//THE OPEN GROUP DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
//IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT
//SHALL TOG BE LIABLE FOR ANY SPECIAL, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
//WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT,
//NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
//PERFORMANCE OF THIS SOFTWARE. 

package ORG.opengroup.axtp.tp;
import ORG.opengroup.axtp.*;
import ORG.opengroup.jar.*;
import ORG.opengroup.axtp.transaction.*;
import java.io.*;
import java.net.*;
import java.util.*;

public class TransactionServerImpl implements TransactionServer
{
	public final int ACTIVE=0;
	public final int ABORTED=1;
	public final int COMMITTED=2;
	private int state=ACTIVE;
	private InetAddress clientAddress;
	private int clientPort;
	private byte[] tid;
	private boolean primary=true;
	private long expiration=0;
	private RetryDelay retryDelay=null;
	private Hashtable ht=null;
	private TransactionObject last=null;
	private long lastTime=0;

	private transient TP tp=null;

	public void setClientAddress(InetAddress clientAddress)
	{
		this.clientAddress=clientAddress;
	}

	public InetAddress getClientAddress()
	{
		return clientAddress;
	}

	public void setClientPort(int clientPort)
	{
		this.clientPort=clientPort;
	}

	public int getClientPort()
	{
		return clientPort;
	}

	public void setTID(byte[] tid)
	{
		this.tid=tid;
	}

	public byte[] getTID()
	{
		return tid;
	}

	public void setPrimary(boolean primary)
	{
		this.primary=primary;
	}

	public boolean getPrimary()
	{
		return primary;
	}

	public boolean isPrimary()
	{
		return primary;
	}

	public void init(TP tp)
	{
		this.tp=tp;
		state=ACTIVE;
		retryDelay=tp.cloneRetryDelay();
		retryDelay.resetDelay();
		ht=new Hashtable();
	}

	public void setDuration(long v)
	{
		if (isPrimary())
			expiration=tp.setTimer(this,TP.DURATION_TIMER,v);
	}

	public long getExpiration()
	{
		return expiration;
	}

	public synchronized void apply(TransactionObject to)
	{
		switch(state)
		{
			case ACTIVE:
			{
				if (to instanceof X)
				{
					Long xid=null;
					ExecutionContext xc=null;
					X x=(X)to;
					long zid=x.getZID();
					if (zid>0)
					{
						xid=new Long(zid);
						xc=(ExecutionContext)ht.get(xid);
						if (xc!=null)
						{
							xc.z();
						}
					}
					xid=new Long(x.getXID());
					xc=(ExecutionContext)ht.get(xid);
					if (xc==null)
					{
						try
						{
							xc=tp.createServerXC(x,this);
							ht.put(xid,xc);
						} catch (Exception e)
						{
							xc=null;
						}
					}
					if (xc!=null)
						xc.apply(x);
				}
				else if (to instanceof C)
				{
					C c=(C)to;
					state=COMMITTED;
					last=getTransactionFactory().createQ();
					lastTime=System.currentTimeMillis();
					sendToClient(last);
					Enumeration enum=ht.elements();
					while(enum.hasMoreElements())
					{
						ExecutionContext xc=(ExecutionContext)enum.nextElement();
						xc.apply(c);
					}
					ht=null;
					retryDelay=null;
				}
				else if (to instanceof R)
				{
					R r=(R)to;
					state=ABORTED;
					last=getTransactionFactory().createW();
					lastTime=System.currentTimeMillis();
					sendToClient(last);
					Enumeration enum=ht.elements();
					while(enum.hasMoreElements())
					{
						ExecutionContext xc=(ExecutionContext)enum.nextElement();
						xc.apply(r);
					}
					ht=null;
					retryDelay=null;
				}
				else if (to instanceof Z)
				{
					Z z=(Z)to;
					Long xid=new Long(z.getXID());
					ExecutionContext xc=(ExecutionContext)ht.get(xid);
					if (xc!=null)
						xc.apply(z);
				}
				return;
			}

			case ABORTED:
			{
				if (to instanceof ClientDecision)
				{
					long tm=System.currentTimeMillis();
					if ((tm-lastTime)>(retryDelay.getInitialDelay()/2))
					{
						lastTime=tm;
						sendToClient(last);
					}
				}
				return;
			}

			case COMMITTED:
			{
				if (to instanceof C)
				{
					long tm=System.currentTimeMillis();
					if ((tm-lastTime)>(retryDelay.getInitialDelay()/2))
					{
						lastTime=tm;
						sendToClient(last);
					}
				}
				return;
			}
		}
	}

	public synchronized void timeout(long context)
	{
		if (context!=TP.DURATION_TIMER)
		{
			if (ht==null)
				return;
			Long xid=new Long(context);
			ExecutionContext xc=(ExecutionContext)ht.get(xid);
			if (xc!=null)
				xc.timeout();
		}
		else if(state==ACTIVE)
		{
			state=ABORTED;
			last=getTransactionFactory().createW();
			sendToClient(last);
			ht=null;
			retryDelay=null;
		}
	}

	public void sendToClient(TransactionObject to)
	{
		to.setAddress(clientAddress);
		to.setPort(clientPort);
		to.setTID(tid);
		tp.send(to);
	}

	public void send(PacketObject po)
	{
		tp.send(po);
	}

	public long setTimer(ExecutionContext xc, long delay)
	{
		return tp.setTimer(this,xc.getXID(),delay);
	}

	public void clearTimer(ExecutionContext xc)
	{
		tp.clearTimer(this,xc.getXID());
	}

	public RetryDelay cloneRetryDelay()
	{
		return (RetryDelay)retryDelay.clone();
	}

	public TransactionFactory getTransactionFactory()
	{
		return tp.getTransactionFactory();
	}

	public boolean isNoisy()
	{
		return tp.isNoisy();
	}
}
